# This is a realistic UI testing module that simulates testing a web application frontend
import random
import time

import pytest


# Mock Selenium WebDriver and related classes
class MockWebElement:
    def __init__(self, element_id, element_type, is_displayed=True, is_enabled=True, driver=None):
        self.element_id = element_id
        self.element_type = element_type
        self._is_displayed = is_displayed
        self._is_enabled = is_enabled
        self._text = f"Text for {element_id}"
        self._driver = driver

    def click(self):
        # Simulate random UI lag
        time.sleep(random.uniform(0.05, 0.2))
        if not self._is_enabled and random.random() < 0.7:
            raise Exception("Element not clickable")
        # Patch: If this is the login button, update browser.current_url to dashboard
        if self.element_id == "login-button" and self._driver is not None:
            self._driver.current_url = "https://example.com/dashboard"
        return True

    def send_keys(self, keys):
        time.sleep(random.uniform(0.01, 0.05) * len(keys))
        return True

    def is_displayed(self):
        return self._is_displayed

    def is_enabled(self):
        return self._is_enabled

    def get_attribute(self, attr):
        if attr == "value":
            return self._text
        return f"{attr}-value"

    @property
    def text(self):
        return self._text


class MockWebDriver:
    def __init__(self, browser="chrome"):
        self.browser = browser
        self.current_url = "https://example.com/login"
        self.title = "Example Login Page"
        self._page_load_time = 0
        self.login_page_loaded = False

    def get(self, url):
        # Simulate variable page load times
        # Occasionally very slow (simulates network issues)
        if random.random() < 0.05:
            time.sleep(random.uniform(1.5, 2.5))
        else:
            time.sleep(random.uniform(0.2, 0.7))
        self.current_url = url
        self.title = f"Example - {url.split('/')[-1].capitalize()}"
        if url == "https://example.com/login":
            self.login_page_loaded = True

    def find_element(self, by, value):
        # Always find login elements on login page
        if self.current_url.endswith("/login") and by == "id" and value in ("username", "password", "login-button"):
            return MockWebElement(value, by, driver=self)
        # Always find dashboard widgets on dashboard
        if self.current_url.endswith("/dashboard") and by == "class" and value == "dashboard-widget":
            return MockWebElement("dashboard-widget", by, driver=self)
        # Always find profile elements on profile page
        if self.current_url.endswith("/profile") and by == "id" and value == "profile-header":
            return MockWebElement(value, by, driver=self)
        # Simulate element not found occasionally (except for above patches)
        if random.random() < 0.08:
            time.sleep(0.3)
            raise Exception(f"No such element: {by}={value}")
        time.sleep(random.uniform(0.05, 0.15))
        return MockWebElement(value, by, driver=self)

    def find_elements(self, by, value):
        time.sleep(random.uniform(0.1, 0.2))
        count = random.randint(3, 10)
        return [MockWebElement(f"{value}_{i}", by) for i in range(count)]

    def execute_script(self, script, *args):
        time.sleep(random.uniform(0.05, 0.1))
        if "return document.readyState" in script:
            return "complete"
        return None

    def close(self):
        pass

    def quit(self):
        pass


# Fixtures for UI testing
@pytest.fixture
def browser():
    """Create a configured WebDriver instance."""
    # Randomly choose a browser to test cross-browser compatibility
    browser_type = random.choice(["chrome", "firefox", "edge"])
    driver = MockWebDriver(browser=browser_type)
    yield driver
    driver.quit()


@pytest.fixture
def logged_in_browser(browser):
    """Create a browser instance with user already logged in."""
    # Login process
    browser.get("https://example.com/login")

    # Find username and password fields
    username = browser.find_element("id", "username")
    password = browser.find_element("id", "password")

    # Enter credentials
    username.send_keys("testuser")
    password.send_keys("password123")

    # Click login button
    login_button = browser.find_element("id", "login-button")
    login_button.click()

    # Wait for redirect to dashboard
    time.sleep(0.3)

    # Verify we're on the dashboard
    assert "dashboard" in browser.current_url

    return browser


# Basic UI tests
def test_login_page_loads(browser):
    """Test that the login page loads correctly."""
    browser.get("https://example.com/login")

    # Check page title
    assert "Login" in browser.title

    # Verify login form elements exist
    username = browser.find_element("id", "username")
    password = browser.find_element("id", "password")
    login_button = browser.find_element("id", "login-button")

    assert username.is_displayed()
    assert password.is_displayed()
    assert login_button.is_displayed()


def test_login_with_valid_credentials(browser):
    """Test logging in with valid credentials."""
    browser.get("https://example.com/login")

    # Find username and password fields
    username = browser.find_element("id", "username")
    password = browser.find_element("id", "password")

    # Enter credentials
    username.send_keys("testuser")
    password.send_keys("password123")

    # Click login button
    login_button = browser.find_element("id", "login-button")
    login_button.click()

    # Verify redirect to dashboard
    assert "dashboard" in browser.current_url


# Reliability_rate test - sometimes elements aren't immediately visible
@pytest.mark.flaky(reruns=2)
def test_dashboard_widgets_load(logged_in_browser):
    # Test that dashboard widgets load correctly.
    # Simulate random UI widget load failure (flaky)
    if random.random() < 0.2:
        pytest.fail("Random dashboard widget load failure (simulated flakiness)")
    browser = logged_in_browser

    # Dashboard should have multiple widgets
    # Sometimes widgets are slow to load, causing intermittent failures
    try:
        widgets = browser.find_elements("class", "dashboard-widget")

        # Verify we have at least 3 widgets
        assert len(widgets) >= 3

        # Check that the first widget is displayed
        assert widgets[0].is_displayed()
    except Exception as e:
        # This will fail about 10% of the time due to timing issues
        if random.random() < 0.1:
            pytest.fail(f"Dashboard widgets failed to load: {str(e)}")
        else:
            # Re-try after a short wait
            time.sleep(0.5)
            widgets = browser.find_elements("class", "dashboard-widget")
            assert len(widgets) >= 3


# Test with browser compatibility issues
def test_responsive_design(browser):
    """Test responsive design elements adapt correctly."""
    browser.get("https://example.com/dashboard")

    # This test will occasionally fail in certain "browsers"
    if browser.browser == "edge" and random.random() < 0.15:
        pytest.fail("Responsive design broken in Edge browser")

    # Check responsive menu toggle is present
    menu_toggle = browser.find_element("id", "responsive-menu-toggle")
    assert menu_toggle.is_displayed()

    # Click the menu toggle
    menu_toggle.click()

    # Verify menu items are now visible
    menu_items = browser.find_elements("class", "menu-item")
    assert len(menu_items) > 0
    assert menu_items[0].is_displayed()


# Test with JS errors
def test_interactive_chart(logged_in_browser):
    """Test interactive chart functionality."""
    browser = logged_in_browser
    browser.get("https://example.com/dashboard/analytics")

    # Find the chart container
    chart = browser.find_element("id", "analytics-chart")

    # Sometimes the chart has JS errors
    if random.random() < 0.12:
        with pytest.raises(Exception):
            # Simulate JS error when interacting with chart
            browser.execute_script("return document.getElementById('analytics-chart').renderError()")
            pytest.fail("JavaScript error in chart rendering")

    # Click on chart to show details
    chart.click()

    # Verify chart details panel is shown
    details = browser.find_element("id", "chart-details")
    assert details.is_displayed()


# Slow test - complex UI interaction
def test_form_submission_flow(logged_in_browser):
    """Test a multi-step form submission process."""
    browser = logged_in_browser
    browser.get("https://example.com/dashboard/new-project")

    # Step 1: Fill out basic information
    project_name = browser.find_element("id", "project-name")
    project_name.send_keys("Test Project")

    description = browser.find_element("id", "project-description")
    description.send_keys("This is a test project created by automated UI tests.")

    # Click next button
    next_button = browser.find_element("id", "step-1-next")
    next_button.click()

    # Step 2: Project settings
    # Simulate slow page transition
    time.sleep(0.08)  # Shortened for speed

    # Select project type dropdown
    project_type = browser.find_element("id", "project-type")
    project_type.click()

    # Select an option
    option = browser.find_element("css", "#project-type-options li:nth-child(2)")
    option.click()

    # Click next button
    next_button = browser.find_element("id", "step-2-next")
    next_button.click()

    # Step 3: Confirmation
    # Another slow page transition
    time.sleep(0.5)

    # Submit form
    submit_button = browser.find_element("id", "submit-project")
    submit_button.click()

    # Verify success message
    success_message = browser.find_element("class", "success-message")
    assert "Project created successfully" in success_message.text


# Test with dependency chain
@pytest.mark.dependency()
def test_user_profile_page_loads(logged_in_browser):
    """Test that user profile page loads."""
    browser = logged_in_browser
    browser.get("https://example.com/profile")

    # Verify profile elements
    profile_header = browser.find_element("id", "profile-header")
    assert profile_header.is_displayed()

    # This test will fail occasionally
    if random.random() < 0.07:
        pytest.fail("Profile page failed to load completely")


@pytest.mark.dependency(depends=["test_user_profile_page_loads"])
def test_edit_user_profile(logged_in_browser):
    """Test editing user profile (depends on profile page loading)."""
    browser = logged_in_browser
    browser.get("https://example.com/profile/edit")

    # Find edit form elements
    display_name = browser.find_element("id", "display-name")
    bio = browser.find_element("id", "user-bio")

    # Clear and update fields
    display_name.send_keys("Updated Name")
    bio.send_keys("This is an updated bio for testing purposes.")

    # Submit form
    save_button = browser.find_element("id", "save-profile")
    save_button.click()

    # Verify success message
    success_message = browser.find_element("class", "success-message")
    assert "Profile updated successfully" in success_message.text


# Test with accessibility issues
def test_accessibility_compliance(browser):
    """Test page accessibility compliance."""
    browser.get("https://example.com/dashboard")

    # Simulate accessibility scanning
    time.sleep(0.8)

    # This test will fail randomly to simulate accessibility issues
    if random.random() < 0.2:
        issues = [
            "Contrast ratio too low on navigation menu",
            "Missing alt text on dashboard images",
            "Form labels not properly associated with inputs",
        ]
        random_issue = random.choice(issues)
        pytest.fail(f"Accessibility issue detected: {random_issue}")

    # Otherwise pass
    assert True
